#!/usr/bin/env python
"""
git-flow

.. program:: git flow

.. cmdoption:: -v, --verbose

       Produce more output.

.. cmdoption:: -h, --help

       Print usage, help and information on the available commands.

"""
#
# This file is part of `gitflow`.
# Copyright (c) 2010-2011 Vincent Driessen
# Copyright (c) 2012 Hartmut Goebel
# Distributed under a BSD-like license. For full terms see the file LICENSE.txt
#

import re
import sys
import argparse
import subprocess as sub

from gitflow.core import GitFlow, info, GitCommandError
from gitflow.util import itersubclasses
from gitflow.jenkins import Jenkins, DeploymentRequestError
from gitflow.exceptions import (GitflowError, AlreadyInitialized,
                                NotInitialized, BranchTypeExistsError,
                                BaseNotOnBranch, NoSuchBranchError,
                                BaseNotAllowed, BranchExistsError,
                                IllegalVersionFormat, InconsistencyDetected,
                                OperationsError)
import gitflow.pivotal as pivotal
import gitflow.review as review
from gitflow.review import (BranchReview, ReviewNotAcceptedYet,
                            get_feature_ancestor)

__copyright__ = "2010-2011 Vincent Driessen; 2012 Hartmut Goebel"
__license__ = "BSD"

def die(*texts):
    raise SystemExit('\n'.join(map(str, texts)))

class NotEmpty(argparse.Action):
    def __call__(self, parser, namespace, values, option_string=None):
        if not values:
            raise argparse.ArgumentError(self, 'must not by empty.')
        setattr(namespace, self.dest, values)


class GitFlowCommand(object):
    """
    This is just an empty class to serve as the base class for all command line
    level sub commands.  Since the git-flow loader will auto-detect all
    subclasses, implementing a new subcommand is as easy as subclassing the
    :class:`GitFlowCommand`.
    """
    @classmethod
    def register_parser(cls, parent):
        raise NotImplementedError("Implement this method in a subclass.")

    @staticmethod
    def run(args):
        raise NotImplementedError("Implement this method in a subclass.")


class VersionCommand(GitFlowCommand):
    @classmethod
    def register_parser(cls, parent):
        p = parent.add_parser('version', help='Show the version of gitflow.')
        p.set_defaults(func=cls.run)

    @staticmethod
    def run(args):
        from gitflow import __version__
        print(__version__)


class StatusCommand(GitFlowCommand):
    @classmethod
    def register_parser(cls, parent):
        p = parent.add_parser('status', help='Show some status.')
        p.set_defaults(func=cls.run)

    @staticmethod
    def run(args):
        gitflow = GitFlow()
        for name, hexsha, is_active_branch in gitflow.status():
            if is_active_branch:
                prefix = '*'
            else:
                prefix = ' '
            info('%s %s: %s' % (prefix, name, hexsha[:7]))


class InitCommand(GitFlowCommand):
    @classmethod
    def register_parser(cls, parent):
        p = parent.add_parser('init',
                              help='Initialize a repository for gitflow.')
        p.add_argument('-f', '--force', action='store_true',
                       help='Force reinitialization of the gitflow preferences.')
        p.add_argument('-d', '--defaults', action='store_true',
                       dest='use_defaults',
                       help='Use default branch naming conventions and prefixes.')
        p.set_defaults(func=cls.run)
        return p

    @staticmethod
    def run(args):
        from . import _init
        _init.run_default(args)


class FeatureCommand(GitFlowCommand):
    @classmethod
    def register_parser(cls, parent):
        p = parent.add_parser('feature', help='Manage your feature branches.')
        sub = p.add_subparsers(title='Actions')
        cls.register_list(sub)
        cls.register_start(sub)
        cls.register_finish(sub)
        cls.register_purge(sub)
        cls.register_checkout(sub)
        cls.register_diff(sub)
        cls.register_rebase(sub)

        cls.register_publish(sub)
        cls.register_pull(sub)
        cls.register_track(sub)

    #- list
    @classmethod
    def register_list(cls, parent):
        p = parent.add_parser('list',
                              help='List all existing feature branches '
                              'in the local and optionally remote repository.')
        p.set_defaults(func=cls.run_list)
        p.add_argument('-a', '--all', action='store_true',
                help='List remote feature branches as well.')
        p.add_argument('-v', '--verbose', action='store_true',
                help='Be verbose (more output).')

    @staticmethod
    def run_list(args):
        gitflow = GitFlow()
        gitflow.start_transaction()
        gitflow.list('feature', 'name', use_tagname=False,
                     verbose=args.verbose, include_remote=args.all)

    #- select
    @classmethod
    def register_start(cls, parent):
        p = parent.add_parser('start', help='Select a story in PT and create'
            'a new feature branch.')
        p.set_defaults(func=cls.run_start)
        p.add_argument('-F', '--no-fetch', action='store_true',
                help='Do not fetch from origin before performing local operation.')
        p.add_argument('-r', '--for-release',
                help='Base the feature branch on a release branch.')
        p.add_argument('-m', '--match',
                help='Filter stories by regular expression')

    @staticmethod
    def run_start(args):
        if args.for_release:
            pivotal.check_version_format(args.for_release)
        gitflow = GitFlow()
        git     = gitflow.git

        base = None
        if args.for_release is not None:
            # Make sure --for-release matches the requirements.
            pivotal.check_version_format(args.for_release)
            base = gitflow.get_prefix('release') + args.for_release
        else:
            base = gitflow.managers['feature'].default_base()

        if not args.no_fetch:
            sys.stderr.write('Fetching origin ... ')
            gitflow.origin().fetch()
            print 'OK'

        # Check if the base exists and is in sync as soon as possible.
        sys.stdout.write('Checking the base branch ({0}) ... '.format(base))
        origin_base = gitflow.require_origin_branch(base)
        try:
            gitflow.must_be_uptodate(base)
        except NoSuchBranchError:
            sys.stdout.write('found remote counterpart ... ')
            git.branch(base, origin_base.name)
        print('OK')

        [story, name] = pivotal.prompt_user_to_select_story(match=args.match)

        sys.stdout.write('Setting myself as the story owner ... ')
        try:
            story.set_me_as_owner()
        except:
            print('FAIL')
        print('OK')

        if args.for_release is not None:
            sys.stdout.write('Assigning the chosen story to release {0} ... '.format(args.for_release))
            story.assign_to_release(args.for_release)
            print('OK')

        if story.is_rejected():
            sid = str(story.get_id())
            gitflow.start_transaction('restart story {0}'.format(sid))
            sys.stdout.write('Checking out the feature branch ... ')
            try:
                gitflow.checkout('feature', sid)
                print('OK')
            except NoSuchBranchError as e:
                print('FAIL')
                raise InconsistencyDetected(
                    'The branch is missing for story {0}.'.format(sid))
            sys.stdout.write('Updating Pivotal Tracker ... ')
            story.start()
            print('OK')
            return

        # :fixme: Why does the sh-version not require a clean working dir?
        gitflow.start_transaction('start feature branch %s (from %s)' % \
                (name, base))
        try:
            # fetch=False because we are already fetching at the beginning.
            branch = gitflow.create('feature', name, base, fetch=False)
        except (NotInitialized, BaseNotOnBranch):
            # printed in main()
            raise
        except Exception, e:
            die("Could not create feature branch %r" % name, e)

        # Mark beginning of the feature branch with another branch
        sys.stdout.write('Inserting feature base marker ... ')
        base_marker = gitflow.managers['feature'].base_marker_name(str(branch))
        git.branch(base_marker, base)
        print('OK')

        sys.stdout.write('Updating Pivotal Tracker ... ')
        story.start()
        print('OK')

        print
        print "Summary of actions:"
        print "- A new branch", branch, "was created, based on", base
        print "- You are now on branch", branch
        print ""
        print "Now, start committing on your feature. When done, use:"
        print ""
        print "     git flow feature finish", name
        print

    #- finish
    @classmethod
    def register_finish(cls, parent):
        p = parent.add_parser('finish', help='Finish a feature branch ' +
            '(with PTintegration).')
        p.set_defaults(func=cls.run_finish)
        #p.add_argument('-F', '--no-fetch', action='store_true', default=True,
        #        help='Fetch from origin before performing local operation.')
        p.add_argument('-r', '--rebase', action='store_true',
                help='Finish branch by rebasing first.')
        #p.add_argument('-k', '--keep', action='store_true', default=True,
        #        help='Keep branch after performing finish.')
        p.add_argument('-D', '--force-delete', action='store_true',
                default=False, help='Force delete feature branch after finish.')
        p.add_argument('-R', '--no-review', action='store_true',
                default=False, help='Do not post a review request.'
                'not just for the last commit.')
        p.add_argument('-S', '--summary-from-commit',
                action='store_true', default=False,
                help='Use the last commit title instead of Pivotal Tracker '
                'story name for summary in Review Board.')
        p.add_argument('-P', '--no-push', action='store_true',
                default=False, help='Do not push the develop branch to origin '
                'after merging with the feature branch.')
        p.add_argument('nameprefix', nargs='?')

    @staticmethod
    def run_finish(args):
        gitflow = GitFlow()
        repo = gitflow.repo
        git = repo.git

        f_prefix = gitflow.get_prefix('feature')
        name = gitflow.nameprefix_or_current('feature', args.nameprefix)
        full_name = f_prefix + name

        #+++ PT story stuff
        sys.stdout.write('Getting data from Pivotal Tracker ... ')
        story_id = pivotal.get_story_id_from_branch_name('feature', name)
        story = pivotal.Story(story_id)
        release = story.get_release()
        print 'OK'

        # Decide on the upstream branch.
        if release:
            # Merge into the release branch.
            r_prefix = gitflow.get_prefix('release')
            short_name = gitflow.nameprefix_or_current('release', release)
            upstream = r_prefix + short_name
        else:
            # Merge into the develop branch.
            upstream = gitflow.develop_name()

        # Fail as soon as possible if something is not right so that we don't
        # get Pivotal Tracker into an inconsistent state.
        rev_range = [get_feature_ancestor(full_name, upstream),
                     repo.commit(full_name).hexsha]

        #+++ Git manipulation
        sys.stdout.write('Finishing feature branch ... upstream %s ... ' \
                         % upstream)
        gitflow.finish('feature', name, upstream=upstream,
                       fetch=True, rebase=args.rebase,
                       keep=True, force_delete=args.force_delete,
                       tagging_info=None, push=(not args.no_push))
        print 'OK'

        #+++ Review Request
        if not args.no_review:
            sys.stdout.write('Posting review ... upstream %s ... ' % upstream)
            review = BranchReview.from_identifier('feature', name, rev_range)
            review.post(story, summary_from_story=(not args.summary_from_commit))
            print('OK')

            sys.stdout.write('Posting code review url into Pivotal Tracker ... ')
            comment = 'New patch was uploaded into Review Board: ' + review.get_url()
            story.add_comment(comment)
            print('OK')

        #+++ Finish PT story
        sys.stdout.write('Updating Pivotal Tracker ... ')
        if story.is_finished():
            sys.stdout.write('story already finished, skipping ... ')
        else:
            sys.stdout.write('finishing %s ... ' % story.get_id())
            story.finish()
        print 'OK'

        #+++ Git modify merge message
        #sys.stdout.write('Amending merge commit message to include links ... ')
        #msg  = 'Finished {0} {1}\n\n'.format(f_prefix, name)
        #msg += 'PT-Story-URL: {0}\n'.format(story.get_url())
        #msg += 'RB-Review-Request-URL: {0}\n'.format(review.get_url())
        #git.commit('--amend', '-m', msg)
        #if not args.no_push:
        #    git.push(gitflow.origin_name(), upstream)
        #print 'OK'

    #- purge
    @classmethod
    def register_purge(cls, parent):
        p = parent.add_parser('purge',
                help='Purge accepted feature branches.')
        p.set_defaults(func=cls.run_purge)

    @staticmethod
    def run_purge(args):
        gitflow = GitFlow()
        git = gitflow.git
        mgr = gitflow.managers['feature']
        origin_name = gitflow.origin_name()

        def prompt_user(question):
            return raw_input(question).strip().lower() == 'y'

        story_cache = {}
        def get_story(story_id):
            try:
                return story_cache[story_id]
            except KeyError:
                sys.stdout.write('Fetching story {0} ... '.format(story_id))
                sys.stdout.flush()
                story = pivotal.Story(story_id)
                story_cache[story_id] = story
                print('')
                return story

        def get_confirmed_branch_set(features, markers):
            to_delete = set()
            # Go through the branches and ask the user.
            for branch in features:
                story_id = pivotal.get_story_id_from_branch_name('feature', branch)
                story = get_story(story_id)
                if not story.is_accepted():
                    continue
                if not prompt_user('Delete {0}? [y/N]: '.format(branch)):
                    continue
                to_delete.add(branch)
                # Check the associated base markers as well.
                base = mgr.base_marker_name(branch)
                if base in markers:
                    to_delete.add(base)
            # Check the markers that were potentially left behind.
            for base in markers:
                if base in to_delete:
                    continue
                story_id = pivotal.get_story_id_from_base_marker(base)
                story = get_story(story_id)
                if not story.is_accepted():
                    continue
                if not prompt_user('Delete {0} (base marker left behind)? [y/N]: '.format(base)):
                    continue
                to_delete.add(base)
            return to_delete

        print('---> Local branches')
        local_branches = [str(b) for b in mgr.iter()]
        local_markers = [str(b) for b in mgr.iter_markers()]
        to_delete = get_confirmed_branch_set(local_branches, local_markers)
        if len(to_delete) != 0:
            print('')
            for branch in to_delete:
                sys.stdout.write('Deleting {} ... '.format(branch))
                sys.stdout.flush()
                try:
                    git.branch('-d', branch)
                    print('OK')
                except Exception as ex:
                    sys.stderr.write('ERR: ' + str(ex) + '\n')
        else:
            print('\nNo local branches selected, skipping...')

        print('\n---> Remote branches')
        remote_branches = [str(b)[len(origin_name)+1:] for b in mgr.iter(remote=True)]
        remote_markers = [str(b)[len(origin_name)+1:] for b in mgr.iter_markers(remote=True)]
        to_delete = get_confirmed_branch_set(remote_branches, remote_markers)
        if len(to_delete) != 0:
            to_push = []
            for branch in to_delete:
                to_push.append(':' + branch)
            sys.stdout.write('\nDeleting the selected remote branches (push) ... ')
            sys.stdout.flush()
            gitflow.origin().push(to_push)
            print('OK')
        else:
            print('\nNo remote branches selected, skipping...')

    #- checkout
    @classmethod
    def register_checkout(cls, parent):
        p = parent.add_parser('checkout',
                help='Check out (switch to) the given feature branch.')
        p.set_defaults(func=cls.run_checkout)
        p.add_argument('nameprefix', action=NotEmpty)

    @staticmethod
    def run_checkout(args):
        gitflow = GitFlow()
        # NB: Does not default to the current branch as `nameprefix` is required
        branch = gitflow.checkout('feature', args.nameprefix)
        print 'Checking out feature {0}.'.format(branch.name)

    #- diff
    @classmethod
    def register_diff(cls, parent):
        p = parent.add_parser('diff',
                help='Show a diff of changes since this feature branched off.')
        p.set_defaults(func=cls.run_diff)
        p.add_argument('nameprefix', nargs='?')

    @staticmethod
    def run_diff(args):
        gitflow = GitFlow()
        name = gitflow.nameprefix_or_current('feature', args.nameprefix)
        gitflow.start_transaction('diff for feature branch %s' % name)
        gitflow.diff('feature', name)

    #- rebase
    @classmethod
    def register_rebase(cls, parent):
        p = parent.add_parser('rebase',
                help='Rebase a feature branch on top of develop.')
        p.set_defaults(func=cls.run_rebase)
        p.add_argument('-i', '--interactive', action='store_true',
                help='Start an interactive rebase.')
        p.add_argument('nameprefix', nargs='?')

    @staticmethod
    def run_rebase(args):
        gitflow = GitFlow()
        name = gitflow.nameprefix_or_current('feature', args.nameprefix)
        gitflow.start_transaction('rebasing feature branch %s' % name)
        gitflow.rebase('feature', name, args.interactive)

    #- publish
    @classmethod
    def register_publish(cls, parent):
        p = parent.add_parser('publish',
                help='Publish this feature branch to origin.')
        p.set_defaults(func=cls.run_publish)
        p.add_argument('nameprefix', nargs='?')

    @staticmethod
    def run_publish(args):
        gitflow = GitFlow()
        git     = gitflow.git
        name = gitflow.nameprefix_or_current('feature', args.nameprefix)
        gitflow.start_transaction('publishing feature branch %s' % name)
        branch = gitflow.publish('feature', name)
        print(branch)
        base_marker = gitflow.managers['feature'].base_marker_name(str(branch))
        git.push(gitflow.origin_name(), base_marker)
        print
        print "Summary of actions:"
        print "- A new remote branch '%s' was created" % branch
        print "- The local branch '%s' was configured to track the remote branch" % branch
        print "- You are now on branch '%s'" % branch
        print

    #- pull
    @classmethod
    def register_pull(cls, parent):
        p = parent.add_parser('pull',
                help='Pull a feature branch from a remote peer.')
        p.set_defaults(func=cls.run_pull)
        p.add_argument('remote', action=NotEmpty,
                       help="Remote repository to pull from.")
        p.add_argument('name', nargs='?',
                help='Name of the feature branch to pull. '
                'Defaults to the current branch, if it is a feature branch.')
        # :todo: implement --prefix
        #p.add-argument('-p', '--prefix',
        #               help='Alternative remote feature branch name prefix.')

    @staticmethod
    def run_pull(args):
        gitflow = GitFlow()
        name = gitflow.name_or_current('feature', args.name, must_exist=False)
        gitflow.start_transaction('pulling remote feature branch %s '
                                  'into local banch %s' % (args.remote, name))
        gitflow.pull('feature', args.remote, name)

    #- track
    @classmethod
    def register_track(cls, parent):
        p = parent.add_parser('track',
                help='Track a feature branch from origin.')
        p.set_defaults(func=cls.run_track)
        p.add_argument('name', action=NotEmpty)

    @staticmethod
    def run_track(args):
        gitflow = GitFlow()
        # NB: `args.name` is required since the branch must not yet exist
        gitflow.start_transaction('tracking remote feature branch %s'
                                  % args.name)
        branch = gitflow.track('feature', args.name)
        print
        print "Summary of actions:"
        print "- A new remote tracking branch '%s' was created" % branch
        print "- You are now on branch '%s'" % branch
        print


class ReleaseCommand(GitFlowCommand):
    @classmethod
    def register_parser(cls, parent):
        p = parent.add_parser('release', help='Manage your release branches.')
        p.add_argument('-v', '--verbose', action='store_true',
           help='Be verbose (more output).')
        sub = p.add_subparsers(title='Actions')
        cls.register_list(sub)
        cls.register_list_stories(sub)
        cls.register_start(sub)
        cls.register_append(sub)
        cls.register_stage(sub)
        cls.register_finish(sub)
        cls.register_track(sub)

    #- list
    @classmethod
    def register_list(cls, parent):
        p = parent.add_parser('list',
                              help='Lists all existing release branches '
                              'in the local repository.')
        p.set_defaults(func=cls.run_list)
        p.add_argument('-v', '--verbose', action='store_true',
                help='Be verbose (more output).')

    @staticmethod
    def run_list(args):
        gitflow = GitFlow()
        gitflow.start_transaction()
        gitflow.list('release', 'version', use_tagname=True,
                     verbose=args.verbose)

    @classmethod
    def register_list_stories(cls, parent):
        p = parent.add_parser('list_stories',
                              help='Lists all stories that are going to be'
                              'released in the release of your choice.')
        p.set_defaults(func=cls.run_list_stories)
        p.add_argument('--version')

    @staticmethod
    def run_list_stories(args):
        print
        if args.version is None:
            pivotal.Release.dump_all_releases()
        else:
            release = pivotal.Release(args.version)
            release.dump_stories()

    #- start
    @classmethod
    def register_start(cls, parent):
        p = parent.add_parser('start', help='Start a new release.')
        p.set_defaults(func=cls.run_start)
        p.add_argument('-F', '--no-fetch', action='store_true',
                help='Do not fetch from origin before performing local operation.')
        p.add_argument('version', action=NotEmpty)

    @staticmethod
    def run_start(args):
        gitflow = GitFlow()
        base = gitflow.develop_name()

        #+ Pivotal Tracker modifications.
        pivotal.prompt_user_to_confirm_release(args.version)

        #+ Git modifications.
        sys.stdout.write('Creating release branch (base being %s) ... ' \
                         % base)
        try:
            branch = gitflow.create('release', args.version, base,
                                    fetch=(not args.no_fetch))
        except BranchExistsError:
            sys.stdout.write('branch already exists ... ')
        except (NotInitialized, BranchTypeExistsError, BaseNotOnBranch):
            # printed in main()
            raise
        except Exception, e:
            die("could not create release branch %r" % args.version, e)
        print 'OK'

        pivotal.start_release(args.version)
        gitflow.checkout('release', args.version)

        print
        print "Follow-up actions:"
        print "- Bump the version number now!"
        print "- Start committing last-minute fixes in preparing your release"
        print "- Wait for all the stories to be QA'd and reviewed"
        print "- When done, run:"
        print
        print "     git flow release stage", args.version
        print
        print "  to push the release to the client staging environment."

    #- append
    @classmethod
    def register_append(cls, parent):
        p = parent.add_parser('append', help='Append stories to an existing release.')
        p.set_defaults(func=cls.run_append)
        p.add_argument('-F', '--no-fetch', action='store_true',
                help='Do not fetch from origin before performing local operation.')
        p.add_argument('version', action=NotEmpty)

    @staticmethod
    def run_append(args):
        # Print info and ask for confirmation.
        pivotal.prompt_user_to_confirm_release(args.version)

        # Merge, push and insert PT labels.
        gitflow = GitFlow()
        git = gitflow.git
        current_branch = gitflow.repo.active_branch

        develop = gitflow.develop_name()
        gitflow.name_or_current('release', args.version)
        release = gitflow.get_prefix('release') + str(args.version)

        print('')
        sys.stdout.write('Merging develop into ' + release + ' ... ')
        gitflow.checkout('release', str(args.version))
        git.merge(gitflow.develop_name())
        print('OK')

        sys.stdout.write('Pushing ' + release + ' ... ')
        gitflow.origin().push([release])
        print('OK')

        sys.stdout.write('Moving back to ' + str(current_branch) + ' ... ')
        current_branch.checkout()
        print('OK')

        sys.stdout.write('Marking Pivotal Tracker stories ... ')
        pivotal.start_release(args.version)
        print('OK')

    #- stage
    @classmethod
    def register_stage(cls, parent):
        p = parent.add_parser('stage', help='Stage a release branch for the client.')
        p.set_defaults(func=cls.run_stage)
        p.add_argument('-F', '--no-fetch', action='store_true',
                help='Do not fetch from origin before performing local operation.')
        p.add_argument('-R', '--ignore-missing-reviews', action='store_true',
                       help='Just print a warning if there is no review for '
                            'a feature branch that is assigned to this release,'
                            ' do not fail.')
        p.add_argument('-D', '--skip-deployment', action='store_true',
                help='Do not deploy to the client staging environment.')
        p.add_argument('version', action=NotEmpty, help="Release to be staged for the client.")

    @staticmethod
    def run_stage(args):
        assert args.version
        pivotal.check_version_format(args.version)

        gitflow = GitFlow()

        # Check the repository if CircleCI is enabled.
        if gitflow.is_circleci_enabled():
            _try_deploy_circleci(gitflow, args.version)

        # Check if all stories were QA'd
        pt_release = pivotal.Release(args.version)
        print('Checking Pivotal Tracker stories ... ')
        pt_release.try_stage()
        print('OK')

        # Check if all relevant review requests are there
        rb_release = review.Release(pt_release)
        print('Checking Review Board review requests ... ')
        rb_release.try_stage(args.ignore_missing_reviews)
        print('OK')

        # Deliver all relevant PT stories
        print('Delivering all relevant Pivotal Tracker stories ... ')
        pt_release.stage()
        print('OK')

        # Trigger the deployment job
        if not args.skip_deployment:
            args.environ = 'client'
            args.no_check = True
            DeployCommand.run_release(args)

        print
        print "Follow-up actions:"
        print "- Wait for the client to accept all the stories."
        print "- When all is done, run"
        print
        print "     git flow release finish", args.version
        print
        print "  to perform the final merge, tagging and cleanup."

    #- finish
    @classmethod
    def register_finish(cls, parent):
        p = parent.add_parser('finish', help='Finish and close a release.')
        p.set_defaults(func=cls.run_finish)
        # fetch by default
        p.add_argument('-F', '--no-fetch', action='store_true',
                help='Do not fetch from origin before performing local operation.')
        # push by default
        p.add_argument('-P', '--no-push', action='store_true',
                       #:todo: get "origin" from config
                       help="Push to origin after performing finish.")
        p.add_argument('-k', '--keep', action='store_true',
                help='Keep branch after performing finish.')
        p.add_argument('-R', '--ignore-missing-reviews', action='store_true',
                       help='Just print a warning if there is no review for '
                            'a feature branch that is assigned to this release,'
                            ' do not fail.')
        p.add_argument('version', action=NotEmpty, help="Release to be finished and closed.")

        g = p.add_argument_group('tagging options')
        g.add_argument('-n', '--notag', action='store_true',
                       help="Don't tag this release.")
        g.add_argument('-m', '--message',
                       help="Use the given tag message.")
        g.add_argument('-s', '--sign', action='store_true',
                help="Sign the release tag cryptographically.")
        g.add_argument('-u', '--signingkey',
                help="Use the given GPG-key for the digital signature "
                     "instead of the default git uses (implies -s).")

    @staticmethod
    def run_finish(args):
        gitflow = GitFlow()
        git     = gitflow.git
        origin  = gitflow.origin()
        version = gitflow.name_or_current('release', args.version)

        #+++ Check if all stories were QA'd
        pt_release = pivotal.Release(args.version)
        print('Checking Pivotal Tracker stories ... ')
        pt_release.try_finish()
        print('OK')

        #+++ Check all relevant review requests in Review Board, to be sure
        rb_release = review.Release(pt_release)
        print('Checking Review Board review requests ... ')
        rb_release.try_finish(args.ignore_missing_reviews)
        print('OK')

        #+++ Merge the release branch into develop and master
        sys.stdout.write('Merging release branch %s ... ' % version)

        # Save the current position of master and develop in case we need to roll back.
        master_hexsha = _branch_hexsha(gitflow.master_name())
        develop_hexsha = _branch_hexsha(gitflow.develop_name())

        tagging_info = None
        if not args.notag:
            tagging_info = dict(
                sign=args.sign or args.signingkey,
                signingkey=args.signingkey,
                message=args.message)
        gitflow.finish('release', version,
                                 fetch=(not args.no_fetch), rebase=False,
                                 keep=True, force_delete=False,
                                 tagging_info=tagging_info, push=False)
        print('OK')

        #+++ Close all relevant review requests
        try:
            sys.stdout.write('Submitting all relevant review requests ... ')
            rb_release.finish()
            print('OK')
        except Exception as ex:
            print('FAIL')
            ReleaseCommand.rollback(tag=args.version,
                    master_hexsha=master_hexsha, develop_hexsha=develop_hexsha)
            raise ex

        #+++ Collect local and remote branches to be deleted
        sys.stdout.write('Collecting branches to be deleted ... ')
        local_branches  = list()
        remote_branches = list()

        #+ Collect features to be deleted.
        origin_prefix = str(origin) + '/'
        feature_prefix = gitflow.get_prefix('feature')
        # refs = [<type>/<id>/...]
        refs = [str(ref)[len(origin_prefix):] for ref in origin.refs]
        for story in pt_release:
            if story.is_rejected():
                continue
            # prefix = <feature-prefix>/<id>
            prefix = feature_prefix + str(story.get_id())
            base_marker = gitflow.managers['feature'].base_marker_name(prefix)
            try:
                name = gitflow.nameprefix_or_current('feature', prefix)
                local_branches.append(feature_prefix + name)
                if base_marker in gitflow.repo.refs:
                    local_branches.append(base_marker)
            except NoSuchBranchError:
                pass
            for ref in refs:
                # if <feature-prefix>/... startswith <feature-prefix>/<id>
                if ref.startswith(prefix) or ref == base_marker:
                    remote_branches.append(ref)
        #+ Collect releases to be deleted.
        if not args.keep:
            release_branch = gitflow.get_prefix('release') + version
            try:
                gitflow.name_or_current('release', version)
                local_branches.append(release_branch)
            except NoSuchBranchError:
                pass
            if release_branch in refs:
                remote_branches.append(release_branch)
            print 'OK'

        #+++ Delete local and remote branches that are a part of this release
        sys.stdout.write('Checking out {0} ...'.format(gitflow.develop_name()))
        try:
            git.checkout(gitflow.develop_name())
        except Exception as ex:
            print('FAIL')
            ReleaseCommand.rollback(tag=args.version,
                    master_hexsha=master_hexsha, develop_hexsha=develop_hexsha)
            raise ex
        print('OK')
        #+ Delete local branches.
        print('Deleting local branches ...')
        try:
            for branch in local_branches:
                print('    ' + branch)
                git.branch('-D', branch)
        except Exception as ex:
            print('FAIL (the last branch printed)')
            ReleaseCommand.rollback(tag=args.version,
                    master_hexsha=master_hexsha, develop_hexsha=develop_hexsha)
            raise ex
        print('    OK')
        #+ Delete remote branches.
        print('Deleting remote branches and pushing the rest ...')
        for branch in remote_branches:
            print('    ' + branch)
        refspecs = [(':' + b) for b in remote_branches]
        refspecs.append(gitflow.develop_name())
        refspecs.append(gitflow.master_name())
        try:
            git.push(str(origin), '--tags', *refspecs)
        except Exception as ex:
            print('    FAIL')
            ReleaseCommand.rollback(tag=args.version,
                    master_hexsha=master_hexsha, develop_hexsha=develop_hexsha)
            raise ex
        print('    OK')

    #- track
    @classmethod
    def register_track(cls, parent):
        p = parent.add_parser('track',
                help='Track a release branch from origin.')
        p.set_defaults(func=cls.run_track)
        p.add_argument('version', action=NotEmpty)

    @staticmethod
    def run_track(args):
        gitflow = GitFlow()
        # NB: `args.version` is required since the branch must not yet exist
        gitflow.start_transaction('tracking remote release branch %s'
                                  % args.version)
        branch = gitflow.track('release', args.version)
        print
        print "Summary of actions:"
        print "- A new remote tracking branch '%s' was created" % branch
        print "- You are now on branch '%s'" % branch
        print

    @staticmethod
    def rollback(tag=None, master_hexsha=None, develop_hexsha=None):
        gitflow = GitFlow()

        try:
            if not tag is None:
                sys.stdout.write('Deleting the release tag ... ')
                gitflow.git.tag('-d', tag)
                print ('OK')
        except Exception:
            print ('FAIL')

        try:
            if not master_hexsha is None:
                sys.stdout.write('Resetting master to the original position ... ')
                _reset_branch(gitflow.master_name(), master_hexsha)
                print('OK')
        except Exception:
            print ('FAIL')

        try:
            if not develop_hexsha is None:
                sys.stdout.write('Resetting develop to the original position ... ')
                _reset_branch(gitflow.develop_name(), develop_hexsha)
                print ('OK')
        except Exception:
            print ('FAIL')


def _branch_hexsha(branch):
    return sub.check_output(['git', 'rev-parse', branch]).replace('\n', '')

def _reset_branch(branch, hexsha):
    current = sub.check_output(['git', 'rev-parse', '--abbrev-ref', 'HEAD']).replace('\n', '')
    sub.check_output(['git', 'checkout', branch], stderr=sub.STDOUT)
    sub.check_output(['git', 'reset', '--keep', hexsha], stderr=sub.STDOUT)
    sub.check_output(['git', 'checkout', current], stderr=sub.STDOUT)


class HotfixCommand(GitFlowCommand):
    @classmethod
    def register_parser(cls, parent):
        p = parent.add_parser('hotfix', help='Manage your hotfix branches.')
        p.add_argument('-v', '--verbose', action='store_true',
           help='Be verbose (more output).')
        sub = p.add_subparsers(title='Actions')
        cls.register_list(sub)
        cls.register_start(sub)
        cls.register_finish(sub)
        cls.register_publish(sub)

    #- list
    @classmethod
    def register_list(cls, parent):
        p = parent.add_parser('list',
                              help='Lists all existing hotfix branches '
                              'in the local repository.')
        p.set_defaults(func=cls.run_list)
        p.add_argument('-v', '--verbose', action='store_true',
                help='Be verbose (more output).')

    @staticmethod
    def run_list(args):
        gitflow = GitFlow()
        gitflow.start_transaction()
        gitflow.list('hotfix', 'version', use_tagname=True,
                     verbose=args.verbose)

    #- start
    @classmethod
    def register_start(cls, parent):
        p = parent.add_parser('start', help='Start a new hotfix branch.')
        p.set_defaults(func=cls.run_start)
        p.add_argument('-F', '--fetch', action='store_true',
                       #:todo: get "origin" from config
                help='Fetch from origin before performing local operation.')
        p.add_argument('version', action=NotEmpty)
        p.add_argument('base', nargs='?')

    @staticmethod
    def run_start(args):
        gitflow = GitFlow()
        # NB: `args.version` is required since the branch must not yet exist
        # :fixme: get default value for `base`
        gitflow.start_transaction('create hotfix branch %s (from %s)' % \
                (args.version, args.base))
        try:
            branch = gitflow.create('hotfix', args.version, args.base,
                                    fetch=args.fetch)
        except (NotInitialized, BranchTypeExistsError, BaseNotOnBranch):
            # printed in main()
            raise
        except Exception, e:
            die("Could not create hotfix branch %r" % args.version, e)
        print
        print "Summary of actions:"
        print "- A new branch", branch, "was created, based on", args.base
        print "- You are now on branch", branch
        print ""
        print "Follow-up actions:"
        print "- Bump the version number now!"
        print "- Start committing your hot fixes"
        print "- When done, run:"
        print
        print "     git flow hotfix finish", args.version

    #- finish
    @classmethod
    def register_finish(cls, parent):
        p = parent.add_parser('finish', help='Finish a hotfix branch.')
        p.set_defaults(func=cls.run_finish)
        p.add_argument('-F', '--fetch', action='store_true',
                help='Fetch from origin before performing local operation.')
        p.add_argument('-p', '--push', action='store_true',
                       #:todo: get "origin" from config
                       help="Push to origin after performing finish.")
        p.add_argument('-k', '--keep', action='store_true',
                help='Keep branch after performing finish.')
        p.add_argument('version', nargs='?')

        g = p.add_argument_group('tagging options')
        g.add_argument('-n', '--notag', action='store_true',
                       help="Don't tag this hotfix.")
        g.add_argument('-m', '--message',
                       help="Use the given tag message.")
        g.add_argument('-s', '--sign', action='store_true',
                help="Sign the hotfix tag cryptographically.")
        g.add_argument('-u', '--signingkey',
                help="Use this given GPG-key for the digital signature "
                     "instead of the default git uses (implies -s).")

    @staticmethod
    def run_finish(args):
        gitflow = GitFlow()
        version = gitflow.name_or_current('hotfix', args.version)
        gitflow.start_transaction('finishing hotfix branch %s' % version)
        tagging_info = None
        if not args.notag:
            tagging_info = dict(
                sign=args.sign or args.signingkey,
                signingkey=args.signingkey,
                message=args.message)
        release = gitflow.finish('hotfix', version,
                                 fetch=args.fetch, rebase=False,
                                 keep=args.keep, force_delete=False,
                                 tagging_info=tagging_info)

    #- publish
    @classmethod
    def register_publish(cls, parent):
        p = parent.add_parser('publish',
                help='Publish this hotfix branch to origin.')
        p.set_defaults(func=cls.run_publish)
        p.add_argument('version', nargs='?')

    @staticmethod
    def run_publish(args):
        gitflow = GitFlow()
        version = gitflow.name_or_current('hotfix', args.version)
        gitflow.start_transaction('publishing hotfix branch %s' % version)
        branch = gitflow.publish('hotfix', version)
        print
        print "Summary of actions:"
        print "- A new remote branch '%s' was created" % branch
        print "- The local branch '%s' was configured to track the remote branch" % branch
        print "- You are now on branch '%s'" % branch
        print


class SupportCommand(GitFlowCommand):
    @classmethod
    def register_parser(cls, parent):
        p = parent.add_parser('support', help='Manage your support branches.')
        p.add_argument('-v', '--verbose', action='store_true',
           help='Be verbose (more output).')
        sub = p.add_subparsers(title='Actions')
        cls.register_list(sub)
        cls.register_start(sub)

    #- list
    @classmethod
    def register_list(cls, parent):
        p = parent.add_parser('list',
                              help='Lists all existing support branches '
                              'in the local repository.')
        p.set_defaults(func=cls.run_list)
        p.add_argument('-v', '--verbose', action='store_true',
                help='Be verbose (more output).')

    @staticmethod
    def run_list(args):
        gitflow = GitFlow()
        gitflow.start_transaction()
        gitflow.list('support', 'version', use_tagname=True,
                     verbose=args.verbose)

    #- start
    @classmethod
    def register_start(cls, parent):
        p = parent.add_parser('start', help='Start a new support branch.')
        p.set_defaults(func=cls.run_start)
        p.add_argument('-F', '--fetch', action='store_true',
                help='Fetch from origin before performing local operation.')
        p.add_argument('name', action=NotEmpty)
        p.add_argument('base', nargs='?')

    @staticmethod
    def run_start(args):
        gitflow = GitFlow()
        # NB: `args.name` is required since the branch must not yet exist
        # :fixme: get default value for `base`
        gitflow.start_transaction('create support branch %s (from %s)' %
                (args.name, args.base))
        try:
            branch = gitflow.create('support', args.name, args.base,
                                    fetch=args.fetch)
        except (NotInitialized, BranchTypeExistsError, BaseNotOnBranch):
            # printed in main()
            raise
        except Exception, e:
            die("Could not create support branch %r" % args.name, e)
        print
        print "Summary of actions:"
        print "- A new branch", branch, "was created, based on", args.base
        print "- You are now on branch", branch
        print ""


class DeployCommand(GitFlowCommand):
    @classmethod
    def register_parser(cls, parent):
        p = parent.add_parser('deploy', help='Deploy a Git branch.')
        p.add_argument('-v', '--verbose', action='store_true',
           help='Be verbose (more output).')
        sub = p.add_subparsers(title='Actions')
        cls.register_develop(sub)
        cls.register_release(sub)
        cls.register_master(sub)

    #- develop
    @classmethod
    def register_develop(cls, parent):
        p = parent.add_parser('develop', help='Deploy the develop branch.')
        p.set_defaults(func=cls.run_develop)

    @staticmethod
    def run_develop(args):
        args.environ = 'develop'
        DeployCommand._run_deploy(args)

    #- release
    @classmethod
    def register_release(cls, parent):
        p = parent.add_parser('release', help='Deploy a release branch.')
        p.set_defaults(func=cls.run_release)
        p.add_argument('-F', '--no-fetch', action='store_true',
                help='Do not fetch from origin before performing local operation.')
        p.add_argument('-R', '--ignore-missing-reviews', action='store_true',
                       help='Just print a warning if there is no review for '
                            'a feature branch that is assigned to this release,'
                            ' do not fail.')
        p.add_argument('-C', '--no-check', action='store_true',
                help="Do not perform any checking before deployment.")
        p.add_argument('version', action=NotEmpty, metavar='VERSION',
                help="Release to be deployed.")
        p.add_argument('environ', action=NotEmpty, metavar='ENVIRON',
                help="Environment to deploy into. " \
                     "Must be one of 'qa' or 'client'.")

    @staticmethod
    def run_release(args):
        assert args.version
        assert args.environ
        pivotal.check_version_format(args.version)

        gitflow = GitFlow()

        # Fetch remote refs.
        if not args.no_fetch:
            sys.stderr.write('Fetching origin ... ')
            gitflow.origin().fetch()
            print('OK')

        # Check the environ argument.
        branch = GitFlow().managers['release'].full_name(args.version)
        if args.environ not in ('qa', 'client'):
            raise DeploymentRequestError(branch, args.environ)

        if args.environ == 'client' and not args.no_check:
            #+++ Check if all stories were accepted by the client
            pt_release = pivotal.Release(args.version)
            print('Checking Pivotal Tracker stories ... ')
            pt_release.try_stage()
            print('OK')

            #+++ Check all relevant review requests in Review Board, to be sure
            rb_release = review.Release(pt_release)
            print('Checking if all relevant stories have been reviewed ... ')
            rb_release.try_stage(args.ignore_missing_reviews)
            print('OK')

        DeployCommand._run_deploy(args)

    #- master
    @classmethod
    def register_master(cls, parent):
        p = parent.add_parser('master', help='Deploy the master branch.')
        p.set_defaults(func=cls.run_master)

    @staticmethod
    def run_master(args):
        args.environ = 'production'
        DeployCommand._run_deploy(args)

    #- deploy helper method
    @staticmethod
    def _run_deploy(args):
        assert args.environ
        if args.environ in ('qa', 'client'):
            assert args.version

        gitflow = GitFlow()

        branches = {
                'develop':    gitflow.develop_name(),
                'qa':         gitflow.managers['release'].full_name(args.version),
                'production': gitflow.master_name(),
        }

        if gitflow.is_circleci_enabled():
            branches['client'] = gitflow.client_name()
            _deploy_circleci(gitflow, branches, args.environ)
        else:
            branches['client'] = gitflow.managers['release'].full_name(args.version)
            _deploy_jenkins(gitflow, branches, args.environ)


def _try_deploy_circleci(gitflow, version):
    # The repository must be completely clean for this step to proceed.
    try:
        output = sub.check_output(["git", "status", "--porcelain"])
    except sub.CalledProcessError as ex:
        sys.stderr.write("git status --porcelain failed\n")
        raise OperationsError(ex)

    lines = output.split('\n')
    for line in lines:
        # The only allowed files are these that are untracked. Any other scenario
        # can lead to git reset failing for one reason or another.
        if line == '' or line.startswith('??'):
            continue
        print("The repository is dirty!")
        print("git status --porcelain:")
        print(output)
        print("Commit or stash your changes, then retry.")
        raise SystemExit("Operation canceled.")

    # The release branch must exist.
    gitflow.name_or_current("release", version)

def _deploy_circleci(gitflow, branches, environ):
    repo = gitflow.repo
    git = repo.git

    sys.stdout.write('Pushing branch for CircleCI  ... ')
    sys.stdout.flush()

    # Make sure that the client branch is pointing to the right place.
    if environ == 'client':
        release = branches['qa']
        gitflow.must_be_uptodate(release)
        if gitflow.client_exists():
            current = gitflow.current_branch()
            git.checkout(gitflow.client_name())
            # Use --keep to make sure that local modification are not discarded.
            git.reset('--keep', release)
            git.checkout(current)
        else:
            git.branch(gitflow.client_name(), release)

    # Push the relevant branch to deploy it.
    branch = branches[environ]
    sys.stdout.write('pushing {0} ... '.format(branch))
    sys.stdout.flush()
    gitflow.origin().push([branch])
    print('OK')

def _deploy_jenkins(gitflow, branches, environ):
    # Make sure that the branch being deployed exists in origin.
    branch = branches[environ]
    sys.stderr.write("Checking whether branch '{0}' exists in origin ... " \
                     .format(branch))
    gitflow.require_origin_branch(branch)
    print('OK')

    # Trigger the job.
    jenkins = Jenkins.from_prompt()

    print('Triggering the deployment job (env being {0}) ... job {1} ... ' \
            .format(environ, jenkins.get_deploy_job_name(environ)))
    cause = 'Triggered by ' + gitflow.get('user.name') + \
        ' using the GitFlow plugin'
    url = jenkins.get_url_for_next_invocation(environ)
    invocation = jenkins.trigger_deploy_job(environ, cause)
    print('OK')

    print('\nThe job has been enqueued. You can visit\n\n\t{0}\n\n' \
            'to see the progress.'.format(url))


def main():
    parser = argparse.ArgumentParser(prog='git flow')
    placeholder = parser.add_subparsers(title='Subcommands')
    for cls in itersubclasses(GitFlowCommand):
        cls.register_parser(placeholder)
    args = parser.parse_args()
    try:
        args.func(args)
    except KeyboardInterrupt:
        raise SystemExit('Aborted by user request.')


if __name__ == '__main__':
    try:
        main()
    except (GitflowError, GitCommandError), e:
        raise SystemExit('Error: %s' %e)



from gitflow.core import GitFlow
import httplib2
import busyflow.pivotal as pt
import string
from colorama import Style
from colorama import init
init()
import sys
import re
import itertools
import json

from gitflow.exceptions import (NotInitialized, GitflowError,
                                ReleaseAlreadyAssigned, IllegalVersionFormat,
                                StatusError, NoSuchBranchError, PointMeError)


PT_V5_ENDPOINT = 'https://www.pivotaltracker.com/services/v5'

_gitflow = GitFlow()

def check_version_format(version):
    matcher = _get_version_matcher()
    if not re.match(matcher + '$', version):
        raise IllegalVersionFormat(matcher)

def _get_version_matcher():
    return _gitflow._safe_get('gitflow.release.versionmatcher')

def _get_token():
    return _gitflow._safe_get('gitflow.pt.token')

def get_client():
    return pt.PivotalClient(token=_gitflow._safe_get('gitflow.pt.token'))

def _get_project_id():
    return _gitflow._safe_get('gitflow.pt.projectid')

def iter_current_stories():
    client = get_client()
    pid = _get_project_id()
    iterations = client.iterations.current(pid)
    if 'iterations' not in iterations:
        return
    for iteration in iterations['iterations']:
        for story in iteration['stories']:
            s = Story.from_dict(story)
            if s.is_feature() or s.is_bug():
                yield s

def iter_backlog_stories():
    client = get_client()
    pid = _get_project_id()
    iterations = client.iterations.backlog(pid)
    if 'iterations' not in iterations:
        return
    for iteration in iterations['iterations']:
        for story in iteration['stories']:
            s = Story.from_dict(story)
            if s.is_feature() or s.is_bug():
                yield s

def iter_stories():
    # Load the PT include/exclude labels from git config.
    # Use string.lower() since PT labels are case insensitive.
    include = _gitflow.get('gitflow.pt.includelabel', None)
    if include is not None:
        include = include.lower()

    exclude = _gitflow.get('gitflow.pt.excludelabels', None)
    if exclude is not None:
        exclude = exclude.lower()
        exclude = exclude.split(',')

    for s in itertools.chain(iter_current_stories(), iter_backlog_stories()):
        if include and not s.is_labeled(include):
            continue
        if exclude and any([l for l in exclude if s.is_labeled(l)]):
            continue
        yield s

def list_projects():
    projects = get_client().projects.all()['projects']
    return [(p['id'], p['name']) for p in projects]

class Story(object):
    def __init__(self, story_id, _skip_story_download=False):
        self._project_id = _get_project_id()
        self._client = get_client()
        if _skip_story_download:
            return
        payload = self._client.stories.get(self._project_id, story_id)
        self._story = payload['story']

    def get_id(self):
        return self._story['id']

    def get_name(self):
        return self._story['name']

    def get_url(self):
        return self._story['url']

    def get_type(self):
        return self._story['story_type']

    def get_state(self):
        return self._story['current_state']

    def set_state(self, state):
        self._update(current_state=state)

    def get_labels(self):
        return self._story.get('labels', [])

    def add_label(self, label):
        labels = self.get_labels()
        if label in labels:
            return
        labels.append(label)
        self._update(labels=labels)

    def is_labeled(self, label):
        return label in self.get_labels()

    def add_comment(self, comment):
        self._client.stories.add_comment(
            project_id=self._project_id, story_id=self.get_id(), text=comment)

    def start(self):
        self.set_state('started')

    def is_started(self):
        return self.get_state() == 'started'

    def is_unstarted(self):
        return self.get_state() == 'unstarted' \
            or self.get_state() == 'unscheduled' \
            or self.get_state() == 'planned'

    def set_me_as_owner(self):
        headers = {'X-TrackerToken': _get_token()}

        # Download /me
        endpoint = PT_V5_ENDPOINT + '/me'
        resp, content = httplib2.Http().request(
                endpoint,
                headers=headers)
        if resp.status != 200:
            raise OperationsError('failed to get Pivotal Tracker /me')

        # Decode my ID
        my_id = None
        try:
            my_id = json.loads(content)['id']
        except ValueError:
            raise OperationsError('failed to decode Pivotal Tracker /me response')

        # Set my ID as the owner of this story
        endpoint = '{0}/projects/{1}/stories/{2}' \
                   .format(PT_V5_ENDPOINT, _get_project_id(), self.get_id())
        req_content = json.dumps({'owned_by_id': my_id})
        headers['Content-Type'] = 'application/json'
        rep, rep_content = httplib2.Http().request(
                endpoint,
                body=req_content,
                method='PUT',
                headers=headers)
        if rep.status != 200:
            raise OperationsError('failed to set me as the story owner')

    #+++ Bug- & Feature-specific stuff
    def finish(self):
        assert self.is_feature() or self.is_bug()
        if self.is_finished():
            return
        if not self.is_started():
            raise StatusError('Feature not started: ' + str(self.get_id()))
        self.set_state('finished')

    def is_finished(self):
        assert self.is_feature() or self.is_bug()
        return self.get_state() == 'finished'

    def deliver(self):
        assert self.is_feature() or self.is_bug()
        if self.is_delivered():
            return
        if not self.is_finished():
            raise StatusError('Feature not finished: ' + str(self.get_id()))
        self.set_state('delivered')

    def is_delivered(self):
        assert self.is_feature() or self.is_bug()
        return self.get_state() == 'delivered'

    def is_accepted(self):
        assert self.is_feature() or self.is_bug()
        return self.get_state() == 'accepted'

    def is_rejected(self):
        assert self.is_feature() or self.is_bug()
        return self.get_state() == 'rejected'

    def get_release(self):
        assert self.is_feature() or self.is_bug()
        for label in self.get_labels():
            m = re.match('release-(' + _get_version_matcher() + ')$', label)
            if m:
                return m.groups()[0]

    def assign_to_release(self, release):
        assert self.is_feature() or self.is_bug()
        check_version_format(release)
        if self.get_release():
            raise ReleaseAlreadyAssigned('Story already assigned to a release')
        self.add_label('release-' + release)

    def get_branch(self):
        assert self.is_feature() or self.is_bug()
        mgr = _gitflow.managers['feature']
        try:
            return mgr.by_name_prefix(str(self.get_id()))
        except NoSuchBranchError:
            pass
        try:
            return mgr.by_name_prefix(str(self.get_id()), remote=True)
        except NoSuchBranchError:
            pass
        return None
    #--- Bug- & Feature-specific stuff

    #+++ Feature-specific stuff
    def is_feature(self):
        return self.get_type() == 'feature'

    def get_estimate(self):
        assert self.is_feature()
        return self._story['estimate']

    def is_estimated(self):
        assert self.is_feature()
        return self.get_estimate() != -1
    #--- Feature-specific stuff

    #+++ Bug-specific stuff
    def is_bug(self):
        return self.get_type() == 'bug'
    #--- Bug-specific stuff

    def dump(self, highlight_labels=[]):
        story = self._story
        sys.stdout.write(colorize_string(story['name']))
        labels = []
        for label in story.get('labels', []):
            if label in highlight_labels:
                lblstr = Style.BRIGHT + str(label) + Style.RESET_ALL
            else:
                lblstr = str(label)
            labels.append(lblstr)
        sys.stdout.write(' ' + Style.DIM + ','.join(labels) + Style.RESET_ALL)
        sys.stdout.write(' (' + Style.DIM + str(story['current_state']) +
            Style.RESET_ALL + ')')
        sys.stdout.write('\n')

    def _update(self, **kwargs):
        try:
            self._client.stories.update(project_id=self._project_id,
                                        story_id=self.get_id(),
                                        **kwargs)
        except pt.RequestError, e:
            msg  = e.parsed_body['message']
            msg += '\n\nMake sure that you are allowed to update this story!'
            raise GitflowError(msg)
        # Commit changes into our internal story instance as well.
        self._story.update(kwargs)

    def to_dict(self):
        return self._story

    @classmethod
    def from_dict(cls, story):
        t = type('Story', (cls,), dict(_story=story))
        return t(0, _skip_story_download=True)


class Release(object):
    def __init__(self, version, _skip_story_download=False):
        check_version_format(version)
        self._version = version

        # Load the PT include/exclude labels from git config.
        # Use string.lower() since PT labels are case insensitive.
        self._include_label = _gitflow.get('gitflow.pt.includelabel', None)
        if self._include_label is not None:
            self._include_label = self._include_label.lower()

        self._exclude_labels = _gitflow.get('gitflow.pt.excludelables', None)
        if self._exclude_labels is not None:
            self._exclude_labels = self._exclude_labels.lower()
            self._exclude_labels = self._exclude_labels.split(',')

        if _skip_story_download:
            return
        self._current_stories = list(iter_current_stories())

    def __iter__(self):
        for story in self._current_stories:
            if story.is_labeled('release-' + self._version):
                if self._include_label:
                    if story.is_labeled(self._include_label):
                        # If the lable is defined, pick only the stories labeled
                        # with that particular lable.
                        yield story
                    continue
                if self._exclude_labels and \
                        any([l for l in self._exclude_labels if story.is_labeled(l)]):
                    # If the story contains any of the exclude labels, skip it.
                    continue
                # The include label is not defined and the story is not matching
                # any of the exclude labels, if defined, so here we can just
                # safely yield the story.
                yield story

    def get_version(self):
        assert self._version
        return self._version

    def start(self):
        for story in self.iter_candidates():
            story.assign_to_release(self._version)

    def try_stage(self):
        err = False
        for story in self:
            if story.is_labeled('point me'):
                err = True
                print("    Story {0} labeled 'point me'!".format(story.get_url()))

            label = None
            for l in ('dupe', 'wontfix', 'cannot reproduce'):
                if story.is_labeled(l):
                    label = l
                    print("    Story {0} labeled '{1}', skipping...".format(story.get_url(), l))
                    break
            if label is not None:
                continue

            if not story.is_labeled('qa+'):
                # Allow zero-point stories not to be QA'd since they are
                # by definition refactoring stories.
                if story.is_feature() and story.get_estimate() == 0:
                    continue
                err = True
                print("    Story {0} not QA'd!".format(story.get_url()))
        if err:
            raise StatusError("Pivotal Tracker check did not pass, operation canceled.")

    def stage(self):
        print 'Following stories were staged for client acceptance as of release %s:' \
              % self._version
        for story in self:
            if not story.is_delivered() \
                    and not story.is_accepted() \
                    and not story.is_rejected():
                story.deliver()
            sys.stdout.write('    ')
            story.dump()

    def try_finish(self):
        self.try_stage()
        err = False
        for story in self:
            if not story.is_accepted():
                err = True
                print('    Story not accepted: ' + story.get_url())
        if err:
            raise StatusError("Pivotal Tracker check did not pass, operation canceled.")

    def prompt_for_confirmation(self):
        answer = raw_input("Do you wish to start the release? [y/N]: ")
        if answer.lower() == 'y':
            return True
        return False

    def iter_candidates(self):
        for story in self._current_stories:
            if self._include_label and not story.is_labeled(self._include_label):
                # Not labeled with the include label, skip it.
                continue
            if self._exclude_labels and \
                    any([l for l in self._exclude_labels if story.is_labeled(l)]):
                # Labeled with an exclude label, skip it.
                continue
            if (story.is_feature() or story.is_bug()) \
                    and story.is_finished() \
                    and story.get_release() is None:
                yield story

    def dump_stories(self):
        print '%s %s %s %s' % (32 * '-', 'Release', self._version, 33 * '-')
        for story in self:
            story.dump(highlight_labels=['release-' + self._version])
        print 80 * '-' + '\n'

    @classmethod
    def dump_all_releases(cls):
        stories = dict()
        for story in iter_current_stories():
            if story.is_feature() or story.is_bug():
                release = story.get_release()
                if release:
                    if release in stories:
                        stories[release].append(story)
                    else:
                        stories[release] = [story]
        for release in stories:
            cls.from_dict(release, stories[release]).dump_stories()

    @classmethod
    def from_dict(cls, version, stories):
        t = type('Release', (cls,), dict(_current_stories=stories))
        return t(version, _skip_story_download=True)


def print_story(story, index=None, highlight_labels=[]):
    if index:
        sys.stdout.write(Style.BRIGHT + str(index) + ' : ' + Style.RESET_ALL)
    sys.stdout.write(colorize_string(story['name']))
    labels = []
    for label in story.get('labels', []):
        if label in highlight_labels:
            lblstr = Style.BRIGHT + str(label) + Style.RESET_ALL
        else:
            lblstr = str(label)
        labels.append(lblstr)
    sys.stdout.write(' ' + Style.DIM + ','.join(labels) + Style.RESET_ALL)
    sys.stdout.write(' (' + Style.DIM + str(story['current_state']) +
        Style.RESET_ALL + ')')
    sys.stdout.write('\n')


def filter_stories(stories, states, types=None):
    types = types or ['feature', 'chore', 'bug']
    return [s for s in stories if s['current_state'] in states and
        s['story_type'] in types]


def prompt_user_to_select_story(match=None):
    i = 1
    stories = list()
    pattern = None
    if match is not None:
        try:
            pattern = re.compile(match)
        except Exception as ex:
            raise GitflowError('failed to parse --match flag: {0}'.format(ex))
    try:
        page_size = int(_gitflow.get('gitflow.pagination'))
    except Exception:
        page_size = 10

    index = ""
    print Style.DIM + "--------- Stories -----------" + Style.RESET_ALL
    for story in iter_stories():
        if story.is_feature() and not story.is_estimated():
            continue
        # You do not start a story if its branch is present or it is
        # assigned to a release, unless it is rejected.
        # If it is not rejected, use checkout instead.
        if (story.get_release() is not None \
                or story.get_branch() is not None) \
                and not story.is_rejected():
            continue
        # Make sure the story state is one of the accepted states.
        if not story.is_started() \
                and not story.is_unstarted() \
                and not story.is_rejected():
            continue
        # If the matching regexp is defined, make sure the story matches.
        if pattern is not None and pattern.search(story.get_name()) is None:
            continue
        # The story passes all the criterias, add it to the list.
        stories.append(story)
        print_story(story.to_dict(), i)

        if i % page_size == 0:
            msg = "\nSelect a story, or press ENTER to see more stories, or type 'q' to quit: "
            index = raw_input(msg)
            if index != "":
                break
            print("")

        i += 1
    print Style.DIM + "-----------------------------" + Style.RESET_ALL
    print

    if len(stories) == 0:
        raise SystemExit('No story available, aborting!\n\n' \
                'Try "git flow feature list" to find another story' \
                ' to work on.')

    # Prompt user to choose story index.
    if index == "":
        print("Select a story, or type 'q' to quit: ")
        index = raw_input()
    if index == "q":
        raise SystemExit("Operation canceled.")

    # We expect a number.
    try:
        index = int(index)
    except ValueError:
        raise GitflowError("Expected a number.")

    if not 0 < index <= len(stories):
        raise GitflowError("Invalid story index: %s." % index)

    # We're expecting input to start from 1, so we have to
    # subtract one here to get the list index.
    story = stories[index - 1]

    if story.is_rejected():
        return [story, None]

    slug = prompt_user_to_select_slug(story)
    return [story, str(story.get_id()) + '/' + slug]


def prompt_user_to_select_slug(story):
    print ("Please choose short story description (will be slugified " +
        "and used as part of the branch name). Don't use whitespace.")
    slug_hint = slugify(get_story_bold_part(story.get_name()))
    print "Story description (default: '%s'):" % slug_hint
    slug = raw_input().strip() or slug_hint
    if any(c in (string.whitespace + '/') for c in slug):
        print ("\n%sERROR%s: The slug must not contain whitespace or '/'.\n" %
            (Style.BRIGHT, Style.RESET_ALL))
        return prompt_user_to_select_slug(story)
    return slug


def get_story_bold_part(story_name):
    """
        >>> get_story_bold_part("*This is the bold part*, this is not.")
        'This is the bold part'
    """
    if not '*' in story_name:
        return story_name
    else:
        return re.sub(r'.*\*(.+)\*.*', r'\1', story_name)


def slugify(story_name, max_length=25):
    """
        >>> slugify(' Just a story name ')
        'just-a-story-name'
        >>> slugify('Just_a_story_name')
        'just_a_story_name'
        >>> slugify('Just_a_story_name_but_a_very_long_one_indeed...')
        'just_a_story_name_but_a_'
        >>> slugify('Just a story name but a very long one indeed...')
        'just-a-story-name-but-a'
    """
    # Inspired by Django.
    value = re.sub('[^\w\s-]', '', story_name).strip().lower()
    value = re.sub('[-\s]+', '-', value)
    if len(value) > max_length:
        value = value[:max_length]
        return value[:(value.rfind('-'))]
    else:
        return value


def get_iterations():
    token = _gitflow.get('gitflow.pt.token')
    project_id = _gitflow.get('gitflow.pt.projectid')
    client = pt.PivotalClient(token=token)
    current = client.iterations.current(project_id)
    backlog = client.iterations.backlog(project_id)
    return [current, backlog]


def update_story(story_id, **kwargs):
    token = _gitflow.get('gitflow.pt.token')
    project_id = _gitflow.get('gitflow.pt.projectid')
    client = pt.PivotalClient(token=token)
    try:
        client.stories.update(
            project_id=project_id, story_id=story_id, **kwargs)
    except pt.RequestError, e:
        msg  = e.parsed_body['message']
        msg += '\n\nMake sure that you are allowed to update this story!'
        raise GitflowError(msg)


def add_comment_to_story(story_id, msg):
    token = _gitflow.get('gitflow.pt.token')
    project_id = _gitflow.get('gitflow.pt.projectid')
    client = pt.PivotalClient(token=token)
    client.stories.add_comment(
        project_id=project_id, story_id=story_id, text =msg)


def get_story(story_id):
    token = _gitflow.get('gitflow.pt.token')
    project_id = _gitflow.get('gitflow.pt.projectid')
    client = pt.PivotalClient(token=token)
    return client.stories.get(project_id=project_id, story_id=story_id)


def finish_story(story_id):
    story = get_story(story_id)
    update_story(story_id, current_state='finished')


def colorize_string(string):
    """ Use ANSI color codes to emulate a simple subset of textile formatting.
        We're supporting '*' (bold).

        Note: Requires 'termcolor' Python module. If not installed, this
        function just returns the string passed in.
    """
    # Remove escaped parts.
    tmp = re.sub('==.*==', '', string)
    bolds = re.findall('\*.+\*', tmp)
    for bold in bolds:
        stripped_asterisks = (Style.BRIGHT + bold.replace('*', '') +
            Style.RESET_ALL)
        tmp = tmp.replace(bold, stripped_asterisks)
    return tmp


def get_story_id_from_branch_name(identifier, branch_name):
    if branch_name.startswith(identifier):
        name = branch_name[len(identifier)+1:]
    else:
        name = branch_name
    match = re.match('^([0-9]+)[/-]?.*$', name)
    if not match:
        raise GitflowError('Weird branch name format: %s' % branch_name)
    return match.groups()[0]


BASE_MARKER_PREFIX = 'base_feature/'

def get_story_id_from_base_marker(base):
    return base[len(BASE_MARKER_PREFIX):]

def show_release_summary(gitflow):
    current = get_iterations()[0]
    current_stories = [
        story for i in current['iterations'] for story in i['stories']]

    print Style.BRIGHT + 'Releasable stories:' + Style.RESET_ALL
    candidate_chores = filter_stories(
        current_stories, ['started'], types=['chore'])
    releasable_chores = [chore for chore in candidate_chores
        if 'reviewed' in chore.get('labels', [])]
    candidate_stories = filter_stories(
        current_stories, ['finished'], types=['bug', 'feature'])
    releasable_stories = [s for s in candidate_stories
        if set(['reviewed', 'qa+']).issubset(set(s.get('labels', [])))]
    for item in releasable_stories + releasable_chores:
        sys.stdout.write(' > ')
        print_story(item)

    print Style.BRIGHT + 'Unreleasable stories:' + Style.RESET_ALL
    unreleasable = [item for item in current_stories
        if not item in releasable_chores and not item in releasable_stories]
    for item in unreleasable:
        sys.stdout.write(' < ')
        print_story(item, highlight_labels=['qa+', 'reviewed'])

    #TODO(Tom): Better release summary based on what's already been merged...
    #lines = gitflow.git.log(
    #    '--show-notes=workflow', '--oneline', '--format="%H %N"').split('\n')
    #finished_stories = []
    #for line in lines:
    #    parts = line.split(' ')
    #    if len(parts) == 3:
    #        (hash, op, story_id) = parts
    #        if op == 'finish':
    #            finished_stories.append((hash, story_id))

    #lines = gitflow.git.log('--oneline', '--format="%h %s"').split('\n')
    ##merges = [l for l in lines if l.split(' ')[1].lower() == 'merge']
    #merged_branches = []
    #for line in lines:
    #    match = re.match(
    #        "(.+)\ Merge\ branch\ '(.+)'\ into %s" % gitflow.develop_name(),
    #        line)
    #    if match:
    #        (hash, branch) = match.groups()
    #        if not branch in merged_branches:
    #            merged_branches.append(branch)
    #print merged_branches
    #releasable = releasable_stories + releasable_chores
    #for story in releasable:
    #    stories_merged_before_this_story = \
    #        set([get_story_id_from_branch_name(s) for s in \
    #            merged_branches[merged_branches.index(story) + 1 : ]])
    #    blockers = stories_merged_before_this_story.intersection(
    #        [s['id'] for s in unreleasable])
    #    if blockers:
    #        print "Story '%s' is blocked by stories: %s" % (story, blockers)
    #import ipdb; ipdb.set_tracce()


def prompt_user_to_confirm_release(version):
    release = Release(version)
    any_assigned = False
    print
    print 'Stories already assigned to this release:'
    for story in release:
        sys.stdout.write('    ')
        story.dump()
        any_assigned = True
    if not any_assigned:
        print '    None'
    print
    any_candidate = False
    any_pointme = False
    print 'Stories to be newly assigned to this release:'
    for story in release.iter_candidates():
        if story.is_labeled('point me'):
            sys.stdout.write('PM  ')
            any_pointme = True
        else:
            sys.stdout.write('    ')
        story.dump()
        any_candidate = True
    if not any_candidate:
        print '    None'
    print

    if not any_candidate:
        raise SystemExit('No new stories to be added to the release,' \
                'aborting...')

    if any_pointme:
        raise PointMeError("Some stories are labeled with the 'point me' label")

    if not release.prompt_for_confirmation():
        raise SystemExit('Aborting...')

def start_release(version):
    Release(version).start()



import ConfigParser
import getpass
import sys

from .core import GitFlow, MAGIC_STRING

def ask(option, question, set_globally=False, secret=False, is_valid=None,
        reuse_existing=True):

    gitflow = GitFlow()
    git = gitflow.repo.git

    answer = None
    try:
        if not reuse_existing:
            raise ConfigParser.NoOptionError(option, 'Not using the existing value')

        answer = gitflow.get(option)
        if isinstance(answer, basestring):
            answer = answer.replace(MAGIC_STRING, '-')
    except (ConfigParser.NoSectionError, ConfigParser.NoOptionError):
        while True:
            if secret:
                answer = getpass.getpass(question)
            else:
                answer = raw_input(question)
            if is_valid is None or is_valid(answer):
                break

        raw_answer = answer.replace('-', MAGIC_STRING)
        if set_globally:
            git.config('--global', option, raw_answer)
        else:
            gitflow.set(option, raw_answer)
    return answer

def pick(option, title, source, reuse_existing=True):
    # Try to get the option from config first.
    gitflow = GitFlow()
    try:
        if not reuse_existing:
            raise ConfigParser.NoOptionError(option, 'Not using the existing value')

        opt = gitflow.get(option)
        if isinstance(opt, basestring):
            opt = opt.replace(MAGIC_STRING, '-')
        return opt
    except (ConfigParser.NoSectionError, ConfigParser.NoOptionError):
        pass

    # Get the suggestions list.
    suggestions = source()
    if len(suggestions) == 0:
        raise SystemExit('No {0} available, exiting...'.format(title))

    # If that fails, ask the user to pick up one of the values.
    print
    print 'Please choose one of the following %s:' % title
    msg = '    Loading...'
    sys.stdout.write(msg)
    sys.stdout.flush()
    sys.stdout.write('\r' * len(msg))

    answer = None
    while not answer:
        i = 0
        for sid, sname in suggestions:
            i += 1
            print '    [%d] %s' % (i, sname)
        inpt = raw_input("Insert the sequence number (or 'q' to quit): ")
        if inpt == 'q':
            raise SystemExit('Operation canceled.')
        try:
            a = int(inpt)
        except ValueError:
            print 'Please specify a number betweet 1 and %i:' % i
            continue
        if a >= 1 and a <= i:
            answer = suggestions[a-1][0]

    value = answer
    if isinstance(answer, basestring):
        value = answer.replace('-', MAGIC_STRING)
    gitflow.set(option, value)
    return answer



import ConfigParser
import getpass
import jenkinsapi.jenkins
import requests
import signal
import sys
import urlparse

from .core import GitFlow, requires_initialized
from .prompt import ask, pick
from .exceptions import (ObjectError, GitflowError)


class DeployJobNotFoundError(ObjectError):
    def __str__(self):
        return 'Jenkins job {0} not found'.format(self.args[0])

class DeploymentRequestError(GitflowError):
    def __str__(self):
        return "You cannot deploy branch {0[0]} into the {0[1]} environment" \
                .format(self.args)

class Jenkins(object):
    def __init__(self, username, password):
        assert username
        assert password
        self._G = GitFlow()
        self._J = jenkinsapi.jenkins.Jenkins(self._get_jenkins_url(),
                username, password)

    def trigger_deploy_job(self, environ, cause=None):
        return self._get_deploy_job(environ).invoke(
                securitytoken=self._get_deploy_job_token(environ), cause=cause)

    def _get_jenkins_url(self):
        prev_handler = None

        def handler(signum, frame):
            try:
                self._G.delete('gitflow.jenkins.url')
                prev_handler(signum, frame)
            except:
                pass

        prev_handler = signal.signal(signal.SIGINT, handler)

        def is_valid(value):
            try:
                return requests.get(value).status_code != 404
            except Exception as ex:
                sys.stderr.write(str(ex) + '\n\n')
                return False

        try:
            value = ask('gitflow.jenkins.url',
                        'Insert the Jenkins server url: ',
                        set_globally=True, is_valid=is_valid)
        except EOFError:
            raise SystemExit('Operation canceled')

        signal.signal(signal.SIGINT, prev_handler)
        return value

    def _get_deploy_job(self, environ):
        job_name = self.get_deploy_job_name(environ)
        if job_name in self._J:
            return self._J[job_name]
        raise DeployJobNotFoundError(job_name)

    def get_deploy_job_name(self, environ):
        return pick('gitflow.jenkins.deployjobname-' + environ,
                'Jenkins jobs as the deploy job',
                lambda: [(k, k) for k in self._J.keys()])

    def _get_deploy_job_token(self, environ):
        req = 'Insert the security token for Jenkins job {0}: ' \
              .format(self.get_deploy_job_name(environ))
        raw = ask('gitflow.jenkins.deployjobtoken-' + environ, req, secret=True)

    def get_url_for_next_invocation(self, environ):
        prefix = self._get_jenkins_url()
        if prefix[-1] != '/':
            prefix += '/'
        job = self._get_deploy_job(environ)
        job_name = self.get_deploy_job_name(environ)
        build_number = job.get_next_build_number()
        return urlparse.urljoin(prefix, 'job/{0}/{1}/'.format(job_name, build_number))

    @classmethod
    def from_prompt(cls):
        username = None
        password = None
        while username is None or username == '':
            username = raw_input('Jenkins username: ')
        while password is None or password == '':
            password = getpass.getpass('Jenkins password: ')
        return cls(username, password)



import re
import subprocess as sub
import difflib as diff
import gitflow.pivotal as pivotal
import reviewboard.extensions as rb_ext
import gitflow.core as core
import sys

from gitflow.core import GitFlow
from gitflow.exceptions import (GitflowError, MultipleReviewRequestsForBranch,
                                NoSuchBranchError, AncestorNotFound, EmptyDiff,
                                PostReviewError, SubmitReviewError)

class ReviewRequestLimitError(GitflowError):
    def __str__(self):
        return 'Too many code review requests: {0}'.format(self.args[0])

class ReviewNotAcceptedYet(GitflowError): pass


_gitflow = GitFlow()

def _get_repo_id():
    return _gitflow._safe_get('gitflow.rb.repoid')

def _get_server():
    return _gitflow._safe_get('reviewboard.server')

def _get_url():
    return _gitflow._safe_get('reviewboard.url')

def _get_client():
    return rb_ext.make_rbclient(_get_server(), '', '')

def _get_develop_name():
    return _gitflow.develop_name()

def _get_branch(identifier, name):
    prefix = _gitflow.get_prefix(identifier)
    name = _gitflow.nameprefix_or_current(identifier, name)
    return prefix + name


def list_repos():
    repos = _get_client().repositories()
    return [(r.id, r.name) for r in repos]


class BranchReview(object):
    def __init__(self, branch, rev_range=None):
        assert rev_range is None or len(rev_range) == 2
        self._branch = branch
        self._client = _get_client()
        if rev_range:
            self._rev_range = rev_range

    def __getattr__(self, name):
        if name == '_rid':
            self._update_from_existing_review()
            return self._rid
        if name == '_summary':
            return None
        if name == '_description':
            return None
        raise AttributeError

    def get_id(self):
        assert self._rid
        return self._rid

    def get_url(self):
        if self._url:
            return self._url
        elif self.get_id():
            return '{0}r/{1}/'.format(_get_url(), self.get_id())
        else:
            raise AttributeError('Neither review id nor review url is defined.')

    def post(self, story, summary_from_story=True):
        assert self._rev_range

        def to_unicode(s):
            try:
                return unicode(s)
            except UnicodeDecodeError:
                return unicode(s, encoding='utf8')

        def to_string(us):
            try:
                return str(us)
            except UnicodeEncodeError:
                return us.encode(encoding='utf8')

        cmd = ['rbt', 'post',
               '--branch', self._branch]

        self._check_for_existing_review()

        desc_cmd = ['git', 'log',
                    "--pretty="
                        "--------------------%n"
                        "Author:    %an <%ae>%n"
                        "Committer: %cn <%ce>%n"
                        "%n"
                        "%s%n%n"
                        "%b",
                    '{0[0]}...{0[1]}'.format(self._rev_range)]
        desc_prefix = u'> Story being reviewed: {0}\n'.format(story.get_url())
        desc = desc_prefix + u'\nCOMMIT LOG\n' + to_unicode(sub.check_output(desc_cmd))

        if summary_from_story:
            summary = story.get_name()
        else:
            # 7 is the magical offset to get the first commit subject
            summary = desc.split('\n')[7]

        # If we are updating an existing review, reuse its summary.
        if self._summary is not None:
            summary = self._summary

        # If we are updating an existing review, reuse part of its description.
        if self._description is not None:
            lines = self._description.split('\n')
            kept_desc = []
            for line in lines:
                if line.startswith('> Story being reviewed'):
                    break
                kept_desc.append(to_unicode(line))
            desc = u'\n'.join(kept_desc) + u'\n' + to_unicode(desc)

        if self._rid:
            sys.stdout.write('updating %s ... ' % str(self._rid))
            cmd.append('--review-request-id')
            cmd.append(str(self._rid))
        else:
            sys.stdout.write('new review ... ')

        cmd.append(u'--summary=' + to_unicode(summary))
        cmd.append(u'--description=' + to_unicode(desc))
        cmd.extend([str(rev) for rev in self._rev_range])
        cmd = [to_string(itm) for itm in cmd]

        p = sub.Popen(cmd, stdout=sub.PIPE, stderr=sub.PIPE)
        (outdata, errdata) = p.communicate()

        if p.returncode == 0:
            print('OK')
            print('>>>> rbt output')
            print(outdata)
            print('<<<< rbt output')
        else:
            debug_cmd = ' '.join(cmd[:-4])
            debug_cmd += " --summary='{0}' --description='{1}' --debug " \
                         .format(summary, desc_prefix)
            debug_cmd += ' '.join(cmd[-2:])
            print('FAIL')
            print('>>>> rbt error output')
            print(errdata)
            print('<<<< rbt error output')
            print('If the error output is not sufficient, execute')
            print('\n    $ {0}\n'.format(debug_cmd))
            print('and see what happens.')
            raise PostReviewError('Failed to post review request.')

        # Use list comprehension to get rid of emply trailing strings.
        self._url = [line for line in outdata.split('\n') if line != ''][-1]
        self._rid = [f for f in self._url.split('/') if f != ''][-1]

    def verify_submit(self):
        assert self._status
        if self._status == 'submitted':
            return
        assert self._rid
        if not self.is_accepted():
            raise ReviewNotAcceptedYet('Review %s has not been accepted yet' \
                                       % str(self._rid))

    def submit(self):
        assert self._status
        if self._status == 'submitted':
            return

        cmd = ['rbt', 'close', '--close-type', 'submitted', str(self.get_id())]
        p = sub.Popen(cmd, stdout=sub.PIPE, stderr=sub.PIPE)
        (outdata, errdata) = p.communicate()

        if p.returncode != 0:
            print('>>>> rbt error output')
            print(errdata)
            print('<<<< rbt error output')
            print('If the error output is not sufficient, execute')
            print('\n    $ rbt close --debug --close-type submitted {0}\n'.format(self.get_id()))
            print('and see what happens.')
            raise SubmitReviewError('Failed to submit review request.')

    def is_accepted(self):
        assert self._rid
        reviews = self._client.get_reviews_for_review_request(self._rid)
        return any(r['ship_it'] for r in reviews)

    def _update(self, **kwargs):
        self._client.update_request(self.get_id(), fields=kwargs, publish=True)

    def _branch_to_rid(self, branch):
        options = dict(repository=_get_repo_id())
        reviews = self._client.get_review_requests(options=options,
                                                   branch=self._branch)
        if len(reviews) > 1:
            raise MultipleReviewRequestsForBranch(self._branch)
        elif len(reviews) == 1:
            r = reviews[0]
            self._summary = r['summary']
            self._description = r['description']
            return r['id']

    def _check_for_existing_review(self):
        assert self._branch
        self._rid = self._branch_to_rid(self._branch)

    @classmethod
    def from_prefix(cls, prefix):
        client = _get_client()
        options = {'repository': _get_repo_id(), 'status': 'pending', 'max-results': 200}
        requests = client.get_review_requests(options=options)
        if len(requests) == 200:
            raise ReviewRequestLimitError(200)

        reviews = [r for r in requests if r['branch'].startswith(prefix) and \
                                          r['status'] != 'discarded']
        if len(reviews) == 0:
            raise NoSuchBranchError(
                    'No review request found for branch prefixed with ' + prefix)
        elif len(reviews) == 1:
            r = reviews[0]
            t = type('BranchReview', (cls,),
                    dict(_rid=r['id'], _status=r['status']))
            return t(r['branch'])
        else:
            raise MultipleReviewRequestsForBranch(reviews[0]['branch'])

    @classmethod
    def from_identifier(cls, identifier, name, rev_range=None):
        prefix = _gitflow.get_prefix(identifier)
        name = _gitflow.nameprefix_or_current(identifier, name)
        return cls(prefix + name, rev_range)


class Release(object):
    def __init__(self, stories):
        self._G = GitFlow()
        self._stories = stories

    def try_stage(self, ignore_missing_reviews):
        assert self._stories
        feature_prefix = self._G.get_prefix('feature')

        self._reviews = []
        reviews_expected = 0
        err = None

        for story in self._stories:
            label = None
            for l in ('no review', 'dupe', 'wontfix', 'cannot reproduce'):
                if story.is_labeled(l):
                    label = l
                    print("    Story {0} labeled '{1}', skipping...".format(story.get_id(), l))
                    break
            if label is not None:
                continue

            prefix = feature_prefix + str(story.get_id())
            try:
                reviews_expected += 1
                review = BranchReview.from_prefix(prefix)
                review.verify_submit()
                print('    ' + str(review.get_id()))
                self._reviews.append(review)
            except (ReviewNotAcceptedYet, NoSuchBranchError) as e:
                print('    ' + str(e))
            except Exception as e:
                print('    ' + str(e))
                err = e
        if err is not None:
            raise SystemExit(str(err))
        if not ignore_missing_reviews and len(self._reviews) != reviews_expected:
            raise SystemExit('Some stories have not been reviewed yet,' \
                    ' aborting...')

    try_finish = try_stage

    def finish(self):
        assert self._reviews is not None
        for review in self._reviews:
            review.submit()

def post_review(self, identifier, name, post_new):
    mgr = self.managers[identifier]
    branch = mgr.by_name_prefix(name)

    sys.stdout.write("Walking the Git reflogs to find review request parent (might "
        "take a couple seconds)...\n")
    sys.stdout.flush()

    if not post_new:
        parent = find_last_patch_parent(self.develop_name(), branch.name)
        if not parent:
            print ("Could not find any merges into %s, using full patch." %
                self.develop_name())
    if not parent:
        parent = get_branch_parent(branch.name)

    if not parent:
        raise GitflowError("Could not find parent for branch '%s'!" %
            branch.name)

    story_id = pivotal.get_story_id_from_branch_name(branch.name)
    story = pivotal.get_story(story_id)

    cmd = ['post-review', '--branch', branch.name,
        '--guess-description',
        '--parent', self.develop_name(),
        '--revision-range', '%s:%s' % (parent, branch.name)]

    if post_new:
        # Create a new request.
        cmd += ['--summary', "'%s'" % story['story']['name']]
    else:
        req = rb_ext.get_latest_review_request_for_branch(
            _gitflow.get('reviewboard.server'), branch.name)
        if req:
            # Update an existing request.
            cmd += ['-r', str(req['id'])]
        else:
            # Create a new request.
            cmd += ['--summary', "'%s'" % story['story']['name']]

    print "Posting a review using command: %s" % ' '.join(cmd)
    proc = sub.Popen(cmd, stdout=sub.PIPE)
    (out, err) = proc.communicate()
    # Post a comment to the relevant Pivotal Tracker story (to make it easier to
    # track review requests).
    review_url = out.strip().split('\n')[-1]
    if not review_url.startswith('http'):
        print ("Could not determine review URL (probably an error when "
            "posting the review")
        return
    if '-r' in cmd:
        comment = "Review request %s updated."
    else:
        comment = "Review request posted: %s"
    pivotal.add_comment_to_story(story_id, comment % review_url)

core.GitFlow.post_review = post_review


def get_branch_parent(branch_name):
    proc = sub.Popen(
        ['git', 'reflog', 'show', branch_name],
        env={'GIT_PAGER': 'cat'}, stdout=sub.PIPE)
    (out, err) = proc.communicate()
    lines = out.strip().split('\n')
    parent = None
    for line in lines:
        parts = line.split(' ')
        if len(parts) >= 3 and (parts[2]).startswith('branch'):
            parent = parts[0]
    if not parent:
        parent = lines[-1].split(' ')[0]
    return parent


def find_last_patch_parent(develop_name, branch_name):
    proc = sub.Popen(
        ['git', 'reflog', 'show', develop_name],
        env={'GIT_PAGER': 'cat'}, stdout=sub.PIPE)
    (out, err) = proc.communicate()
    lines = out.strip().split('\n')
    lines = [l for l in lines if ("merge %s" % branch_name) in l]
    if len(lines) < 2:
        return None
    else:
        # Get the hash of the second most recent merge.
        return lines[1].split(' ')[0]

def get_feature_ancestor(feature, upstream):
    repo = _gitflow.repo

    # Check if we are not looking for the ancestor of the same commit.
    # If that is the case, the algorithm used further fails.
    fc = repo.commit(feature)
    uc = repo.commit(upstream)
    if fc == uc:
        raise EmptyDiff('{0} and {1} are pointing to the same commit.' \
                .format(upstream, feature))

    base_marker = _gitflow.managers['feature'].base_marker_name(feature)
    for ref in repo.refs:
        if str(ref) == base_marker:
            return ref

    raise AncestorNotFound('Base marker missing for ' + feature)
